import inspect
import time
from functools import wraps
from threading import Lock
from typing import Any, Callable, Dict, Mapping, Optional, Type, TypeVar, cast

from .metrics import Metrics
from .resolver import DependencyResolver


class DecoratorResolver:
    def __init__(
        self, resolver: DependencyResolver, metrics: Optional[Metrics] = None, cache_enabled: bool = True
    ) -> None:
        self.__resolver = resolver
        self.__metrics = metrics
        self.__cache_enabled = cache_enabled
        self.__cache: Dict[str, Any] = {}
        self.__lock = Lock()
        self.__parameters: Optional[Mapping[str, inspect.Parameter]] = None

    def preload(self) -> None:
        self.__cache = self.construct_dependencies()

    def clear_cache(self) -> None:
        self.__cache.clear()

    def construct_dependencies(self) -> Dict[str, Any]:
        injected_args: Dict[str, Any] = {}

        if self.__parameters is None:
            raise Exception("parameters not set")

        for parameter_name, parameter in self.__parameters.items():
            default = parameter.default

            if (
                parameter_name == "return"
                or default == inspect.Signature.empty
                or not isinstance(default, ParameterDependence)
            ):
                continue

            dependency: Any = self.__resolver.resolve(default.parameter_dependency).get_instance()
            injected_args[parameter_name] = dependency

        return injected_args

    def inject_function(self, fn: Callable[..., Any]) -> Callable[..., Any]:
        self.__parameters = inspect.signature(fn).parameters

        @wraps(fn)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            t1 = time.perf_counter()
            injected_args: Dict[str, Any] = {}

            if self.__cache_enabled:
                self.__lock.acquire()

                if not self.__cache:
                    self.__cache = self.construct_dependencies()

                self.__lock.release()

                injected_args = self.__cache
            else:
                injected_args = self.construct_dependencies()

            if self.__metrics is not None:
                dt = time.perf_counter() - t1
                self.__metrics.save_metric(fn.__name__, dt)

            return fn(*args, **{**injected_args, **kwargs})

        return wrapper


class ParameterDependence:
    def __init__(self, parameter_dependency: Any) -> None:
        self.parameter_dependency = parameter_dependency


T = TypeVar("T")


def Depends(parameter_type: Type[T]) -> T:
    return cast(T, ParameterDependence(parameter_type))



from typing import Any, Callable, Generic, Protocol, TypeVar

from ..container import Container

T = TypeVar("T")


class Celery(Generic[T], Protocol):
    def task(self, fn: Callable[..., T]) -> Callable[..., T]:
        raise NotImplementedError


def celery_task(celery: Celery[T], container: Container) -> Callable[[Callable[..., Any]], Callable[..., T]]:
    def decorator(fn: Callable[..., Any]) -> Callable[..., T]:
        return celery.task(container.inject(fn))

    return decorator



from typing import Any, Protocol


class Dependency(Protocol):
    def get_instance(self) -> Any:
        raise NotImplementedError


class StaticDependency:
    def __init__(self, instance: Any) -> None:
        self.__instance = instance

    def get_instance(self) -> Any:
        return self.__instance



import inspect
from typing import Any, Callable, Dict, Optional, Protocol, Type, Union, cast, get_type_hints

from pydantic import BaseSettings

from .dependency import Dependency, StaticDependency
from .exceptions import ResolverError
from .scoped_dict import ScopedDict

Dependable = Union[Callable[..., Any], Type[Any]]


class DependencyResolver:
    def __init__(self, container: ScopedDict[Dependable, Dependency]) -> None:
        self._container = container

    def resolve(self, dependency: Dependable, parameters: Optional[Dict[str, Any]] = None) -> Dependency:
        if dependency in self._container:
            return self._container[dependency]

        callable_dependency: Callable[..., Any]
        is_class = False
        if inspect.isclass(dependency):
            if issubclass(cast(type, dependency), BaseSettings):
                return StaticDependency(dependency())

            if issubclass(cast(type, dependency), cast(type, Protocol)):
                raise ResolverError(f"Implementation for {dependency.__name__} protocol is not defined.")

            callable_dependency = dependency.__init__  # type: ignore
            is_class = True
        elif inspect.isfunction(dependency):
            callable_dependency = dependency
        else:
            raise ResolverError(f"Dependency must be a function or a class, {dependency} found.")

        type_hints = get_type_hints(callable_dependency)
        args: Dict[str, Any] = {}

        if parameters:
            parameter_names = type_hints.keys()

            for parameter_name, parameter_value in parameters.items():
                if parameter_name not in parameter_names:
                    raise ResolverError(f"Parameter {parameter_name} it not part of {dependency.__name__}'s signature.")

                args[parameter_name] = parameter_value

        signature_parameters = inspect.signature(callable_dependency).parameters
        for parameter_name, parameter_dependency in type_hints.items():
            if parameter_name == "return" or parameter_name in args:
                continue

            if (default_value := signature_parameters[parameter_name].default) != inspect.Signature.empty:
                args[parameter_name] = default_value
            else:
                try:
                    args[parameter_name] = self.resolve(parameter_dependency).get_instance()
                except ResolverError as e:
                    raise ResolverError(f"{dependency} -> " + str(e))

        skippable_params = (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        injectable_arguments = [p for n, p in signature_parameters.items() if p.kind not in skippable_params]
        number_of_parameters = len(injectable_arguments)

        if is_class:
            # because the callable is init and the self must be skipped
            number_of_parameters -= 1

        if (missing := number_of_parameters - len(args)) > 0:
            raise ResolverError(
                f"Can resolve dependencies for {dependency}. All type annotations must be specified, {missing} missing."
            )

        return StaticDependency(instance=dependency(**args))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['inseminator', 'inseminator.integrations']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'inseminator',
    'version': '0.4.7',
    'description': 'Python dependency injection library based on type hints',
    'long_description': "# Inseminator\n\n[![codecov](https://codecov.io/gh/sukovanej/container/branch/master/graph/badge.svg)](https://codecov.io/gh/sukovanej/container)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n*(definition from dictionary.com)*\n\n> a technician who introduces prepared ~semen~ **dependencies** into the ~genital tract \n> of breeding animals~ python classes, especially ~cows and mares~ pure classes with \n> proper IoC, for ~artificial insemination~ well coupled components and clear classes \n> signatures.\n\nPython library for type-based dependency injection. Write code without global state\nand noisy boilerplate. Inseminator is meant to be used in an entry-point layer of your\napplication and the only thing it requires is properly type-hinted classes \ndependencies.\n\n## Installation\n\nInstall using the pip tool.\n\n```\npip install inseminator\n```\n\n## Usage\n\nYou start by defining the *container* of your dependencies. Whenever you want the container\nto resolve a dependency, it uses the container to search for existing objects and a \nresolver automatically creates desired dependencies.\n\n```Python\nfrom inseminator import Container\n\n\nclass DomainModel:\n    def __init__(self):\n        self.__logic_constant = 1\n\n    def domain_logic(self, input_value: int) -> int:\n        return input_value + self.__logic_constant\n\n\nclass Controller:\n    def __init__(self, domain_model: DomainModel):\n        self.__domain_model = domain_model\n\n    def handler(self, input_value: int) -> int:\n        return self.__domain_model.domain_logic(input_value)\n\n\n# entry-point of your application\n\ncontainer = Container()\n\n# view layer handling\n\ncontroller = container.resolve(Controller)\nresult = controller.handler(1)\nprint(result)\n```\n\nThe strategy for resolving `Controller` is its constructor signature. The resolver works as follows.\n\n  1) We ask the `container` to resolve a dependency `Controller` -> `container.resolve(Controller)`.\n  2) Resolver inside the `container` checks the `Controller`'s constructor signature, i.e. **type hints**\n     of `__init__` method and sees `domain_models: DomainModel`.\n  3) If an instance of `DomainModel` class is already known by the `container` it uses that instance.\n     In the opposite case, the container starts the same resolving machinery for `DomainModel` - which\n     is the exact case we are facing now.\n  4) Because `DomainModel` doesn't have any dependencies it can construct it directly.\n  5) Now the resolver has all the dependencies for `Controller` constructor and can instantiate it.\n\nIf we programmed against an interface instead of implementation the example is modified like this.\n\n```Python\nfrom inseminator import Container\n\nfrom typing import Protocol\n\nclass DomainModel(Protocol):\n    def domain_logic(self, input_value: int) -> int:\n        ...\n\nclass Controller:\n    def __init__(self, domain_model: DomainModel):\n        self.__domain_model = domain_model\n\n    def handler(self, input_value: int) -> int:\n        return self.__domain_model.domain_logic(input_value)\n\n\n# domain model implementation\n\n\nclass ConcreteDomainModel:\n    def __init__(self):\n        self.__logic_constant = 1\n\n    def domain_logic(self, input_value: int) -> int:\n        return input_value + self.__logic_constant\n\n\n# entry point of your application\n\ncontainer = Container()\ncontainer.register(DomainModel, value=ConcreateDomainModel())\n\n# view layer handling\n\ncontroller = container.resolve(Controller)\nresult = controller.handler(1)\nprint(result)\n```\n\nIn this situation, protocol `DomainModel` doesn't hold implementation details, only interface.\nUsing \n\n```\ncontainer.register(DomainModel, value=ConcreateDomainModel())\n```\n\nwe're guiding the resolver to use instance of `ConcreateDomainModel` in case someone asks\nfor `DomainModel`.\n\n### Enforced parameters\n\nIf it is not desired to provide a single concrete implementation for abstract or protocol dependency\none can enforce the resolver to use concrete types for specified parameters. Simply call `container.resolve`\nalso with keywords and tell the resolve how it should resolve some particular parameters.\n\n```Python\ncontainer = Container()\ncontroller = container.resolve(Controller, domain_model=ConcreteDomainModel())\n```\n\nMoreover, using this approach `ConcreteDomainModel` is not evaluated and saved in the container but\nrather in a sub-container which exists only during the resolving. Therefore, if we want to create\nanother instance that depends on `DomainModel` we must either use `register` or again specify\nthe parameter during resolving.\n\n\n### Injecting functions\n\n\nIt might be convinient to specify funcion's dependencies in-place. The great example is Flask\nhandler function. It should live in the same layer the DI container lives because it provides\nonly infrastructure functionality and desirably the only thing it does it calling domain layer's\nfunctions. For this purpose, there is `injector` decorator on the `Container` object. You just\ntell which dependency to provide using `Depends` type constructor.\n\n\n```Python\nfrom inseminator import Container, Depends\n\n\nclass Dependency:\n    def __init__(self):\n        self.x = 1\n\n\ncontainer = Container()\n\n\n@container.inject\ndef my_handler(input_value: int, dependency: Dependency = Depends(Dependency)):\n    return input_value + dependency.x\n```\n\nUsed like that, `my_handler` takes a single argument and thanks to closure it has `dependency` \nprepared with the right instance of `Dependency`.\n\n\n```\n>>> my_handler(1)\n2\n```\n\n### Default parameter values\n\n\nWhen default parameters are specified the resolver uses them unless we override that value\nby enforced parameter.\n\n```python\ndef MyDependency:\n    def __init__(self, parameter: int = 1) -> None:\n        self.parameter = parameter\n\nmy_dependency = container.resolve(MyDependency)\nassert my_dependency.parameter == 1\n```\n\n\n## Docs\n\n - [Clean architecture introduction (draft version)](docs/introduction.md)\n",
    'author': 'Milan Suk',
    'author_email': 'Milansuk@email.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.github.com/sukovanej/inseminator/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
